﻿using GrapeCity.Documents.Pdf;
using System;
using System.Collections.Generic;

using System.IO;
using System.Linq;

using _Float = System.Single;
using _Size = System.Drawing.SizeF;

using GcPdfViewerSupportApi.Models;
using GrapeCity.Documents.Pdf.Annotations;
using GrapeCity.Documents.Pdf.AcroForms;

namespace GcPdfViewerSupportApi.Utils
{
    /// <summary>
    /// PDF document loader.
    /// </summary>
    public class GcPdfDocumentLoader : IDisposable
    {
        #region ** fields

        private static object lockDocumentLoad = new object();
        private _Float _resolution;
        private Stream _stream;
        private GcPdfDocument _doc;

        private DocumentModifications _modifications;
        private SignatureInfo _signatureInfo;
        private SupportApiSettings _settings;
        private DocumentOptions _docOptions;

        #endregion

        #region ** constructor

        /// <summary>
        /// Constructor.
        /// </summary>
        public GcPdfDocumentLoader(SupportApiSettings settings, DocumentOptions docOptions, _Float resolution = 96)
        {
            _settings = settings;
            _docOptions = docOptions;
            _resolution = resolution;            
            Url = "";
            ClientID = docOptions.clientID;
        }

        #endregion

        #region ** properties
        
        public string ClientID { get; private set; }

        public OpenDocumentInfo Info { get; private set; }

        public string Url { get; }

        public int Rotate { get; set; } = 0;

        public _Float Scale { get; set; } = 1.0f;

        public bool RenderInteractiveForms { get; internal set; }

        /// <summary>
        /// Document modifications.
        /// </summary>
        public DocumentModifications DocumentModifications
        {
            get
            {
                return _modifications;
            }
        }

        #endregion

        #region ** methods

        public OpenDocumentInfo Open(Stream data)
        {
            lock (lockDocumentLoad)
            {
                if (_stream != null)
                {
                    _stream.Dispose();
                    _stream = null;
                    _doc = null;
                }
                _stream = data;
                return _OpenInitialDocState();
            }
        }

        public void Reset()
        {
            _modifications = null;
            _signatureInfo = null;
        }

        public void ApplyDocumentModifications(DocumentModifications modifications)
        {
            _modifications = modifications;
            _applyModifications();
        }

        public void Sign(SignatureInfo signatureInfo)
        {
            _signatureInfo = signatureInfo;
        }

        public void Save(MemoryStream s)
        {
            _applyModifications();
            if (_signatureInfo != null)
            {
                _saveWithSignature(s);
            }
            else
            {
                _doc.Save(s);
            }
            s.Seek(0, SeekOrigin.Begin);
        }

        public bool VerifySignature(string fieldName)
        {
            var signatureField = _doc.AcroForm.Fields.Where(f => f.Name == fieldName).FirstOrDefault() as SignatureField;
            if (signatureField == null || signatureField.Value == null)
            {
                return false;
            }
            return signatureField.Value.VerifySignature();
        }

        #endregion

        #region ** IDisposable

        protected virtual void Dispose(bool disposing)
        {
            _doc = null;
            if (_stream != null)
            {
                _stream.Dispose();
                _stream = null;
            }

        }

        public void Dispose()
        {
            // Dispose of unmanaged resources.
            Dispose(true);
            // Suppress finalization.
            GC.SuppressFinalize(this);
        }

        #endregion

        #region ** private and internal implementation

        private void _applyModifications()
        {            
            _OpenInitialDocState();// Reset previous changes.
            List<DataParser> allClientAnnotations = null;
            if (_modifications != null)
            {
                if (_modifications.rotation.HasValue)
                {
                    Rotate = _modifications.rotation.Value;
                }
                if (_modifications.renderInteractiveForms.HasValue)
                {
                    RenderInteractiveForms = _modifications.renderInteractiveForms.Value;
                }
                // Apply document structure changes:                
                var structureChanges = _modifications.structureChanges;
                if (structureChanges != null)
                {
                    bool add = false;
                    int pageIndex = -1;
                    int operationIndex = 0;
                    int checkNumPages = -1;

                    foreach (var change in structureChanges)
                    {
                        add = change.add;
                        pageIndex = change.pageIndex;
                        checkNumPages = change.checkNumPages;

                        if (checkNumPages > 5 && !_doc.IsLicensed)
                        {
                            throw new NotLicensedException("GcPdf is not licensed, and has a limit of 5 loaded pages.");
                        }
                        try
                        {
                            if (checkNumPages != _doc.Pages.Count)
                            {
                                // The most likely reason for this is using an unlicensed GcPdfDocument, which limits the number of pages.
                                // To ignore this error and return whatever pages have been added, comment out the next line:
                                throw new Exception("NumPages check mismatch.");
                            }
                            if (change.add)
                                _doc.Pages.Insert(change.pageIndex);
                            else
                                _doc.Pages.RemoveAt(change.pageIndex);
                            operationIndex++;
                        }
                        catch (Exception ex)
                        {
                            throw new Exception($"Invalid document structure changes. operationIndex:{operationIndex}, pageIndex:{pageIndex}, add:{add}, checkNumPages: {checkNumPages}, doc pages count: {_doc.Pages.Count}. {ex.Message}");
                        }
                    }

                }
                // Apply annotation modifications
                if (_modifications.annotationsData != null)
                {
                    List<KeyValuePair<PopupAnnotation, string>> popupAnnotations = new List<KeyValuePair<PopupAnnotation, string>>();
                    List<KeyValuePair<MarkupAnnotation, string>> referencedAnnotations = new List<KeyValuePair<MarkupAnnotation, string>>();

                    // Update existing annotations:
                    DataParser.CollectAnnotationsData(out List<DataParser> annotationsToUpdate, out List<DataParser> annotationsToCreate,
                        out List<KeyValuePair<int, PdfObjID>> annotationToRemove, out allClientAnnotations, _modifications, _doc, Info);
                    foreach (var dataParser in annotationsToUpdate)
                    {
                        PdfObjID objID = dataParser.pdfObjID;
                        AnnotationTypeCode annotationType = dataParser.annotationType;
                        try
                        {
                            if (dataParser.IsWidget)
                            {
                                Field field = _FindFieldByObjId(_doc.AcroForm.Fields, objID);
                                if (field == null)
                                {
                                    if (dataParser.IsRadioButton)
                                        field = _doc.AcroForm.Fields.Where(a => a.Widgets.Any(w => w.ParsedObjID == objID)).FirstOrDefault();
                                    if (field == null)
                                        continue;
                                }
                                if (dataParser.FieldName != field.Name)
                                {
                                    // DOC-2025
                                    var fieldWidget = field.Widgets.Where(fw => fw.ParsedObjID == objID).FirstOrDefault();
                                    if (fieldWidget != null)
                                    {                                        
                                        if(field.Widgets.Count <= 1)
                                        {
                                            _RemoveField(field);
                                        }
                                        else
                                        {
                                            _RemoveAnnotation(fieldWidget);
                                        }
                                        annotationsToCreate.Add(dataParser);
                                    }
                                    continue;
                                }
                                AcroFormFactory.UpdateField(annotationType, field, dataParser);
                                AddField(field, dataParser);
                            }
                            else
                            {
                                AnnotationBase annotation = dataParser.page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                                if (annotation == null)
                                {
                                    //throw new Exception($"Annotation with id {objID} not found.");
                                    continue;
                                }
                                AnnotationFactory.UpdateAnnotation(annotationType, annotation, dataParser);
                                AddAnnotation(annotation, dataParser.page.Annotations, dataParser);
                                if (annotationType == AnnotationTypeCode.REDACT && dataParser.ParseBool("redactApplied"))
                                {
                                    _doc.Redact(annotation as RedactAnnotation, _settings.RedactOptions);
                                }
                                if (annotationType == AnnotationTypeCode.POPUP)
                                {
                                    popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.parentId));
                                }
                                if ((annotation is MarkupAnnotation) && !string.IsNullOrEmpty(dataParser.referenceAnnotationId))
                                {
                                    referencedAnnotations.Add(new KeyValuePair<MarkupAnnotation, string>((MarkupAnnotation)annotation, dataParser.referenceAnnotationId));
                                }
                            }


                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                    // Remove annotations:
                    foreach (var removedAnnotation in annotationToRemove)
                    {
                        try
                        {
                            Page page = _doc.Pages[removedAnnotation.Key];
                            PdfObjID objID = removedAnnotation.Value;

                            var field = _doc.AcroForm.Fields.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                            if (field != null)
                            {
                                _RemoveField(field);
                            }
                            else
                            {
                                var annotation = page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                                if (annotation != null)
                                {
                                    _RemoveAnnotation(annotation);

                                }

                            }

                            

                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                    // Add new annotations:
                    foreach (var dataParser in annotationsToCreate)
                    {
                        try
                        {
                            var page = dataParser.page;
                            if (dataParser.IsWidget)
                            {
                                if (dataParser.IsRadioButton)
                                {
                                    string fieldName = dataParser.FieldName;
                                    var radioButtonFieldWithSameName = _doc.AcroForm.Fields.Where(a => a.Name == fieldName).FirstOrDefault();
                                    if (radioButtonFieldWithSameName != null)
                                    {
                                        var widget = new WidgetAnnotation(dataParser.page, dataParser.rect);                                        
                                        radioButtonFieldWithSameName.Widgets.Add(widget);
                                        AcroFormFactory.UpdateFieldWidget(radioButtonFieldWithSameName, widget, dataParser);                                        
                                        continue;
                                    }
                                }
                                Field field = AcroFormFactory.CreateField(dataParser, page);
                                if (field != null)
                                {                                    
                                    AddField(field, dataParser);
                                }
                            }
                            else
                            {
                                AnnotationBase annotation = AnnotationFactory.CreateAnnotation(dataParser, page);
                                if (annotation != null)
                                {
                                    AddAnnotation(annotation, page.Annotations, dataParser);
                                    if (dataParser.annotationType == AnnotationTypeCode.REDACT && dataParser.ParseBool("redactApplied"))
                                    {
                                        _doc.Redact(annotation as RedactAnnotation, _settings.RedactOptions);
                                    }
                                    if (dataParser.annotationType == AnnotationTypeCode.POPUP)
                                    {
                                        popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.parentId));
                                    }
                                }
                                if ((annotation is MarkupAnnotation) && !string.IsNullOrEmpty(dataParser.referenceAnnotationId))
                                {
                                    referencedAnnotations.Add(new KeyValuePair<MarkupAnnotation, string>((MarkupAnnotation)annotation, dataParser.referenceAnnotationId));
                                }
                            }
                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                    // Associate popup annotations with their destinations:
                    foreach (var keyValue in popupAnnotations)
                    {
                        try
                        {
                            var popupAnnotation = keyValue.Key;
                            var parentId = keyValue.Value;
                            var parentObjId = DataParser.IdToPdfObjID(parentId);
                            MarkupAnnotation popupParent = popupAnnotation.Page.Annotations.Where(a => a.ParsedObjID == parentObjId).FirstOrDefault() as MarkupAnnotation;
                            if (popupParent == null)
                            {
                                var annotationParser = allClientAnnotations.Where(p => p.id == parentId).FirstOrDefault();
                                if (annotationParser != null)
                                    popupParent = annotationParser.Annotation as MarkupAnnotation;
                            }
                            if (popupParent != null)
                                popupParent.Popup = popupAnnotation;
                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                    // Set text annotation references:
                    foreach (var keyValue in referencedAnnotations)
                    {
                        try
                        {
                            var referenceAnnotation = keyValue.Key;
                            var referenceAnnotationId = keyValue.Value;
                            var referenceAnnotationObjId = DataParser.IdToPdfObjID(referenceAnnotationId);
                            MarkupAnnotation ownerAnnotation = referenceAnnotation.Page.Annotations.Where(a => a.ParsedObjID == referenceAnnotationObjId).FirstOrDefault() as MarkupAnnotation;
                            if (ownerAnnotation == null)
                            {
                                var annotationParser = allClientAnnotations.Where(p => p.id == referenceAnnotationId).FirstOrDefault();
                                if (annotationParser != null)
                                    ownerAnnotation = annotationParser.Annotation as MarkupAnnotation;
                            }
                            if (ownerAnnotation != null)
                            {
                                referenceAnnotation.ReferenceAnnotation = ownerAnnotation;
                                var reviewOrReplyAnnotation = referenceAnnotation as TextAnnotation;
                                if(reviewOrReplyAnnotation != null)  
                                {
                                    if (reviewOrReplyAnnotation.StateModel == "Review")
                                    {
                                        // Review annotation
                                        reviewOrReplyAnnotation.Subject = ownerAnnotation.Subject;
                                        reviewOrReplyAnnotation.ReferenceType = AnnotationReferenceType.Reply;
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Hidden);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Print);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoZoom);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoRotate);
                                        reviewOrReplyAnnotation.CreationDate = new PdfDateTime(DateTime.Now);
                                        reviewOrReplyAnnotation.Rect = ownerAnnotation.Rect;
                                        if (string.IsNullOrEmpty(reviewOrReplyAnnotation.Name))
                                        {
                                            reviewOrReplyAnnotation.Name = Guid.NewGuid().ToString();
                                        }
                                        if (reviewOrReplyAnnotation.StateModel == "Review")
                                        {
                                            reviewOrReplyAnnotation.Contents = $"{reviewOrReplyAnnotation.State} set by {reviewOrReplyAnnotation.UserName}";
                                            reviewOrReplyAnnotation.RichText = $"<?xml version=\"1.0\"?><body xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:xfa=\"http://www.xfa.org/schema/xfa-data/1.0/\" xfa:APIVersion=\"Acrobat:20.6.0\" xfa:spec=\"2.0.2\" ><p>{reviewOrReplyAnnotation.Contents}</p></body>";
                                        }
                                    }
                                    else if(reviewOrReplyAnnotation.ReferenceType == AnnotationReferenceType.Reply)
                                    {
                                        if (string.IsNullOrEmpty(reviewOrReplyAnnotation.StateModel))
                                        {
                                            // Reply annotation:
                                            reviewOrReplyAnnotation.CreationDate = new PdfDateTime(DateTime.Now);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Print);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoZoom);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoRotate);
                                        }
                                    }
                                }
                            }
                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                }
                // Import form data
                if (_modifications.formData != null && _modifications.formData.Count > 0)
                {
                    try
                    {
                        List<KeyValuePair<string, IList<string>>> fieldValues = new List<KeyValuePair<string, IList<string>>>();
                        foreach (var pageIndex in _modifications.formData.Keys)
                        {
                            if (_modifications.formData.TryGetValue(pageIndex, out Dictionary<string, string[]> pageData))
                            {
                                foreach (var kv in pageData)
                                {
                                    fieldValues.Add(new KeyValuePair<string, IList<string>>(kv.Key, new List<string>(kv.Value)));
                                }
                            }
                        }
                        _doc.ImportFormDataFromCollection(fieldValues.ToArray());
                    }
                    catch (Exception)
                    {
                        //
                    }
                }
            }

            var annotationsOrderTable = _modifications.annotationsOrderTable;
            int rotate = Rotate;
            var pages = _doc.Pages;
            foreach (var p in pages)
            {
                p.Rotate = rotate;                
                if (annotationsOrderTable != null && annotationsOrderTable.ContainsKey(p.Index))
                {
                    _SortPageAnnotations(p, new List<string>(annotationsOrderTable[p.Index]), allClientAnnotations);
                }
            }

        }

        private Field _FindFieldByObjId(FieldCollection fields, PdfObjID objID)
        {
            if (fields == null)
                return null;
            foreach (var field in fields)
            {
                if (field.ParsedObjID.Equals(objID))
                    return field;
                var child = _FindFieldByObjId(field.Children, objID);
                if(child  != null)
                    return child;
            }
            return null;
        }

        private Field _FindFieldByName(FieldCollection fields, string fieldName)
        {
            if (fields == null)
                return null;
            foreach (var field in fields)
            {
                if (!string.IsNullOrEmpty(field.Name) && field.Name.Equals(fieldName))
                    return field;
                var child = _FindFieldByName(field.Children, fieldName);
                if (child != null)
                    return child;
            }
            return null;
        }

        private string _FindClientId(AnnotationBase annotation, List<DataParser> allClientAnnotations)
        {
            string clientId = annotation.ParsedObjID.IsEmpty() ? _FindAnnotationId(annotation, allClientAnnotations) : DataParser.PdfObjIDtoId(annotation.ParsedObjID);
            if (string.IsNullOrEmpty(clientId) && annotation is WidgetAnnotation)
            {
                var dataParser = allClientAnnotations.Where(a => a.IsWidget == true && a.Widget != null && a.Widget == annotation).FirstOrDefault();
                if (dataParser == null)
                {
                    var field = (annotation as WidgetAnnotation).Field;
                    dataParser = allClientAnnotations.Where(a => a.IsWidget == true && a.Field == field).FirstOrDefault();
                }
                if (dataParser != null)
                {
                    clientId = dataParser.id;
                }
            }
            return clientId;
        }

        private void _SortPageAnnotations(Page page, List<string> annotationsOrder, List<DataParser> allClientAnnotations)
        {
            AnnotationCollection annotations = page.Annotations;
            if (annotations == null || annotationsOrder == null || allClientAnnotations == null)
                return;
            annotations.Sort(new Comparison<AnnotationBase>(delegate (AnnotationBase x, AnnotationBase y)
            {
                string xID = _FindClientId(x, allClientAnnotations);
                string yID = _FindClientId(y, allClientAnnotations); 
                int xIndex = annotationsOrder.IndexOf(xID);
                int yIndex = annotationsOrder.IndexOf(yID);
                if (xIndex == -1 || yIndex == -1)
                {
                    if(x is WidgetAnnotation && y is WidgetAnnotation)
                    {
                        var xName = ((WidgetAnnotation)x).Field?.Name ?? "";
                        var yName = ((WidgetAnnotation)y).Field?.Name ?? "";
                        return xName.CompareTo(yName);
                    }
                    return -1;
                }
                var res = xIndex.CompareTo(yIndex);
                return res;
            }));
        }
       
        private string _FindAnnotationId(AnnotationBase annotation, IEnumerable<DataParser> allClientAnnotations)
        {
            if (allClientAnnotations == null) 
                return "";
            var fieldParser = allClientAnnotations.Where(p => p.Field != null && p.Field.Equals(annotation)).FirstOrDefault();
            return fieldParser != null ? fieldParser.id : "";
        }

        private void AddField(Field field, DataParser dataParser)
        {
            try
            {
                FieldCollection owner = field.Owner;
                string uniqueFieldName = field.Name;
                // Remove original field from collection. (DOC-2001)
                _RemoveField(field);
                dataParser.Field = field;
                // Then search for duplicate names. (DOC-2001)
                int duplicateCount = 0;                
                Field duplicateField = _FindFieldByName(_doc.AcroForm.Fields, uniqueFieldName);
                while (duplicateField != null)
                {
                    if(duplicateField.Widgets.Count == 0 || duplicateField.Widgets[0].Page == null)
                    {
                        // Fix for DOC-1991: replace field without page.
                        _RemoveField(duplicateField);
                        break;
                    }
                    uniqueFieldName = $"{field.Name}_{++duplicateCount}";
                    duplicateField = _FindFieldByName(_doc.AcroForm.Fields, uniqueFieldName);
                }
                field.Name = uniqueFieldName;

                if(owner == null)
                    owner = _doc.AcroForm.Fields;
                owner.Add(field);
                foreach (var widget in field.Widgets)
                    widget.Page = dataParser.page;
            } 
            catch (Exception)
            {
                throw;
            }
        }

        private void AddAnnotation(AnnotationBase annotation, AnnotationCollection collection, DataParser dataParser)
        {
            _RemoveAnnotation(annotation);
            dataParser.Annotation = annotation;
            collection.Add(annotation);
            annotation.Page = dataParser.page;
        }

        private void _RemoveAnnotation(AnnotationBase annotation)
        {
            foreach (var p in _doc.Pages)
            {
                if (p.Annotations.Contains(annotation))
                {
                    p.Annotations.Remove(annotation);
                    break;
                }
            }
        }

        private void _RemoveField(Field field)
        {
            foreach (var widget in field.Widgets.ToArray())
            {
                _RemoveAnnotation(widget);
            }
            var owner = field.Owner;
            if (owner == null)
                owner = _doc.AcroForm.Fields;
            owner.Remove(field);
        }

        private void _saveWithSignature(Stream stream)
        {
            if (_signatureInfo != null)
            {

                var props = new SignatureProperties();
                props.ContactInfo = _signatureInfo.ContactInfo;
                props.Location = _signatureInfo.Location;
                props.SignerName = _signatureInfo.SignerName;
                props.SigningDateTime = DateTime.Now;
                props.Reason = _signatureInfo.Reason;
                props.SignatureDigestAlgorithm = _signatureInfo.SignatureDigestAlgorithm;
                props.SignatureFormat = _signatureInfo.SignatureFormat;
                props.TimeStamp = _signatureInfo.TimeStamp;
                props.SignatureField = _signatureInfo.SignatureField;
                props.Certificate = _settings.Certificate;
                _doc.Sign(props, stream);
            }
        }

        private OpenDocumentInfo _OpenInitialDocState()
        {
            try
            {
                if (_doc != null)
                {
                    _doc.Clear();
                    _doc = null;
                }
                string password = null;
                if (_docOptions != null)
                    password = _docOptions.password;
                var d = new GcPdfDocument();                
                d.Load(_stream, password);
                _doc = d;
                _Size defaultViewPortSize = d.Pages.Count > 0 ? d.Pages[0].GetRenderSize(this._resolution, this._resolution) : new _Size(0f, 0f);                
                Info = new OpenDocumentInfo(d.DocumentInfo, d.Pages.Count, ClientID, defaultViewPortSize, this._docOptions);
                return Info;
            } 
            catch(Exception ex)
            {
                GcPdfViewerSupportApi.Controllers.GcPdfViewerController.SetLastError(ex.Message);
                throw;
            }
        }

        #endregion

    }
}
